/****************************************************************
 * Licensed to the Apache Software Foundation (ASF) under one   *
 * or more contributor license agreements.  See the NOTICE file *
 * distributed with this work for additional information        *
 * regarding copyright ownership.  The ASF licenses this file   *
 * to you under the Apache License, Version 2.0 (the            *
 * "License"); you may not use this file except in compliance   *
 * with the License.  You may obtain a copy of the License at   *
 *                                                              *
 *   http://www.apache.org/licenses/LICENSE-2.0                 *
 *                                                              *
 * Unless required by applicable law or agreed to in writing,   *
 * software distributed under the License is distributed on an  *
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY       *
 * KIND, either express or implied.  See the License for the    *
 * specific language governing permissions and limitations      *
 * under the License.                                           *
 ****************************************************************/

package org.apache.james.mailetcontainer.camel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import javax.mail.MessagingException;

import org.apache.camel.Body;
import org.apache.camel.Handler;
import org.apache.camel.InOnly;
import org.apache.camel.Property;
import org.apache.commons.logging.Log;
import org.apache.james.core.MailImpl;
import org.apache.james.mailetcontainer.lib.ProcessorUtil;
import org.apache.mailet.Mail;
import org.apache.mailet.MailAddress;
import org.apache.mailet.Matcher;

/**
 * A Splitter for use with Camel to split the MailMessage into many pieces if needed. This is done
 * by use a Matcher.
 * 
 *
 */
@InOnly
public class MatcherSplitter {
    
    /**
     * Headername which is used to indicate that the matcher matched
     */
    public final static String MATCHER_MATCHED_ATTRIBUTE = "matched";
    
    /**
     * Headername under which the matcher is stored
     */
    public final static String MATCHER_PROPERTY = "matcher";

    public final static String ON_MATCH_EXCEPTION_PROPERTY = "onMatchException";
    
    public final static String LOGGER_PROPERTY = "logger";

    
    /**
     * Generate a List of MailMessage instances for the give @Body. This is done by using the given Matcher to see if we need more then one instance of the 
     * MailMessage
     * 
     * @param matcher Matcher to use for splitting
     * @param mail Mail which is stored in the @Body of the MailMessage
     * @return mailMessageList
     * @throws MessagingException
     */
    @SuppressWarnings("unchecked")
    @Handler
    public List<Mail> split(@Property(MATCHER_PROPERTY) Matcher matcher, @Property(ON_MATCH_EXCEPTION_PROPERTY) String onMatchException, @Property(LOGGER_PROPERTY) Log logger, @Body Mail mail) throws MessagingException {
        List<Mail> mails = new ArrayList<Mail>();
        boolean fullMatch = false;
        
        // call the matcher
        Collection<MailAddress> matchedRcpts = null;
        
        try {
            matchedRcpts = matcher.match(mail);
            if (matchedRcpts == null) {
                //In case the matcher returned null, create an empty Collection
                matchedRcpts = new ArrayList<MailAddress>(0);
            } else if (matchedRcpts != mail.getRecipients()) {
                //Make sure all the objects are MailAddress objects
                ProcessorUtil.verifyMailAddresses(matchedRcpts);
            }
        } catch (MessagingException me) {
        
            if (onMatchException == null) {
                onMatchException = Mail.ERROR;
            } else {
                onMatchException = onMatchException.trim().toLowerCase(Locale.US);
            }
            if (onMatchException.compareTo("nomatch") == 0) {
                //In case the matcher returned null, create an empty Collection
                matchedRcpts = new ArrayList<MailAddress>(0);
            } else if (onMatchException.compareTo("matchall") == 0) {
                matchedRcpts = mail.getRecipients();
                // no need to verify addresses
            } else {
                ProcessorUtil.handleException(me, mail, matcher.getMatcherConfig().getMatcherName(), onMatchException, logger);
            }
        }
        
        // check if the matcher matched
        if ( matchedRcpts != null && matchedRcpts.isEmpty() == false) {
            List<MailAddress> rcpts = new ArrayList<MailAddress>(mail.getRecipients());
            
            Iterator<MailAddress> rcptsIterator = matchedRcpts.iterator();
            
            while(rcptsIterator.hasNext()) {
                // loop through the recipients and remove the recipients that matched
                rcpts.remove(rcptsIterator.next());
            }
            
            if (rcpts.isEmpty()) {
                // all recipients matched
                fullMatch = true;
            } else {
                mail.setRecipients(rcpts);
                
                Mail newMail = new MailImpl(mail);
                newMail.setRecipients(matchedRcpts);
                
              
                // Set a header because the matcher matched. This can be used later when processing the route
                newMail.setAttribute(MATCHER_MATCHED_ATTRIBUTE, true);
                
                // add the new generated mail to the mails list
                mails.add(newMail);
            }
        }
            
        if (fullMatch) {
            // Set a header because the matcher matched. This can be used later when processing the route
            mail.setAttribute(MATCHER_MATCHED_ATTRIBUTE, true);
        }
        
        // add mailMsg to the mails list
        mails.add(mail);
        
        return mails;
    }

}
